"""Small utilities that play around with propositions.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

"""
from __future__ import nested_scopes
import operator
import structure
import copy



def getIdentifiers(tree):
    """Return all the unique identifiers of a propositional parse
    tree."""
    children_identifiers = mappend(getIdentifiers, tree.children)
    if isIdentifierNode(tree):
        return uniq([tree.attr] + children_identifiers)
    return uniq(children_identifiers)


def makeFullStateTables(identifiers):
    """Given a list of identifier names, returns a list of all possible
    environments for these identifiers.

    Be careful: this is very expensive and is guaranteed to return
    2**(len(identifiers)) elements!  If I have time, I should probably
    make this an iterative algorithm, or avoid constructing this
    altogether."""
    if len(identifiers) == 0: return [{}]
    partial_state_tables_true = makeFullStateTables(identifiers[1:])
    partial_state_tables_false = copy.deepcopy(partial_state_tables_true)
    for i in range(len(partial_state_tables_true)):
        partial_state_tables_true[i][identifiers[0]] = structure.TRUE_VALUE
        partial_state_tables_false[i][identifiers[0]] = structure.FALSE_VALUE
    return partial_state_tables_true + partial_state_tables_false


def getTrueStates(proposition_tree):
    """Given a proposition_tree, return the list of states such that
    evaluation in any of of them causes the proposition to evaluate
    as a true statement."""
    ids = getIdentifiers(proposition_tree)
    all_states = makeFullStateTables(ids)
    true_states = filter(lambda s: (proposition_tree.eval(s) ==
                                    structure.TRUE_VALUE),
                         all_states)
    return true_states


def getFalseStates(proposition_tree):
    """Given a proposition_tree, return the list of states such that
    evaluation in any of of them causes the proposition to evaluate
    as a false statement."""
    ids = getIdentifiers(proposition_tree)
    all_states = makeFullStateTables(ids)
    false_states = filter(lambda s: (proposition_tree.eval(s) ==
                                    structure.FALSE_VALUE),
                         all_states)
    return false_states

                         

def implicationToDisjunction(impl_node):
    """Given an implication, returns an equivalent disjunction.
    Otherwise, just returns impl_node.

    Not destructive."""
    if not isImplicationNode(impl_node): return impl_node
    return structure.OrNode(structure.NotNode(impl_node[0]),
                            impl_node[1])


def disjunctiveNormalize(prop_tree):
    """Given a proposition, return a new proposition in normalized
    disjunctive form.  This is an answer to Exercise 2.7 of Gries's
    "The Science of Programming"."""
    true_states = getTrueStates(prop_tree)
    def make_mini_prop(state):
        children = []
        for id, value in state.items():
            if value == structure.TRUE_VALUE:
                children.append(structure.IdentifierNode(id))
            else:
                children.append(structure.NotNode(structure.
                                                  IdentifierNode(id)))
        return structure.AndNode(*children)
    mini_props = [make_mini_prop(s) for s in true_states]
    return structure.OrNode(*mini_props)


def conjunctiveNormalize(prop_tree):
    """Given a proposition, returns a new proposition in normalized
    conjunctive form.  This is an answer to Exercise 2.8 of Gries's
    "The Science of Programming"."""
    false_states = getFalseStates(prop_tree)
    def make_mini_prop(state):
        children = []
        for id, value in state.items():
            if value == structure.TRUE_VALUE:
                children.append(structure.IdentifierNode(id))
            else:
                children.append(structure.NotNode(structure.
                                                  IdentifierNode(id)))
        return deMorganize(structure.NotNode(structure.AndNode(*children)))
    mini_props = [make_mini_prop(s) for s in false_states]
    return structure.AndNode(*mini_props)


def deMorganize(not_node):
    """Reduces the node by using De Morgan's law:

    not (a or b) ==> (not a) and (not b)
    not (a and b) ==> (not a) or (not b)
    """
    if not isNotNode(not_node): return not_node
    new_children = map(structure.NotNode, not_node[0].children)
    if isAndNode(not_node[0]):
        return structure.OrNode(*new_children)
    elif isOrNode(not_node[0]):
        return structure.AndNode(*new_children)
    return not_node


def notNotSimplify(not_node):
    '''Given 2 nots in a row, remove both.  In short, not not x ==> x'''
    if not isNotNode(not_node): return not_node
    if not isNotNode(not_node[0]): return not_node
    return not_node[0][0]


def associateAndChains(and_node):
    """Applies the reduction: (and (and x y) z) ==> (and x y z)"""
    if not isAndNode(and_node): return and_node
    and_children, non_and_children = sift(isAndNode, and_node.children)
    new_and_children = mappend(lambda x:x.children, and_children)
    return structure.AndNode(*(new_and_children + non_and_children))


def associateOrChains(or_node):
    """Applies the reduction: (or (or x y) z) ==> (or x y z)"""
    if not isOrNode(or_node): return or_node
    or_children, non_or_children = sift(isOrNode, or_node.children)
    new_or_children = mappend(lambda x:x.children, or_children)
    return structure.OrNode(*(new_or_children + non_or_children))


def orSimplify(or_node):
    """(or T exp) ==> T,
    (or F exp) ==> exp,
    and (or F F) ==> F"""
    if not isOrNode(or_node):
        return or_node
    if filter(isTrueNode, or_node.children):
        return structure.TrueNode()
    false_nodes, other_children = sift(isFalseNode, or_node.children)
    if len(other_children) >= 2:
        return structure.OrNode(*other_children)
    elif len(other_children) == 1:
        return other_children[0]
    else: return structure.FalseNode()
## Perhaps I should have the OrNode constructor handle some of these
## cases.


def andSimplify(and_node):
    """(and T exp) ==> exp"""
    if not isAndNode(and_node):
        return and_node
    if filter(isFalseNode, and_node.children):
        return structure.FalseNode()
    true_nodes, other_children = sift(isTrueNode, and_node.children)
    if len(other_children) >= 2:
        return structure.AndNode(*other_children)
    elif len(other_children) == 1:
        return other_children[0]
    else: return structure.TrueNode()
## Perhaps I should have the AndNode constructor handle this...


def equalityByTruth(prop1, prop2):
    """Return 1 if prop1 and prop2 evaulate to equivalent true states,
    and 0 otherwise."""
    true_states = getTrueStates(prop1), getTrueStates(prop2)
    return sublist(true_states[0], true_states[1]) and \
           sublist(true_states[1], true_states[0])



######################################################################
def isIdentifierNode(node):
    return isinstance(node, structure.IdentifierNode)


def isAndNode(node):
    return isinstance(node, structure.AndNode)


def isOrNode(node):
    return isinstance(node, structure.OrNode)


def isNotNode(node):
    return isinstance(node, structure.NotNode)


def isImplicationNode(node):
    return isinstance(node, structure.ImplicationNode)


def isEqualityNode(node):
    return isinstance(node, structure.EqualityNode)


def isTrueNode(node):
    return isinstance(node, structure.TrueNode)


def isFalseNode(node):
    return isinstance(node, structure.FalseNode)

######################################################################


def mappend(function, l):
    """Perform a mapping append.  Useful when the function's result
    is some kind of list."""
    if not l: return []
    return reduce(operator.add, map(function, l))


def uniq(L):
    """Return all the unique elements of l as a list"""
    dict = {}
    for element in L: dict[element] = 1
    return dict.keys()


def negate(f):
    """Negates a boolean function."""
    return lambda x: not f(x)


def deepTransform(function, tree):
    """Apply some transformation deeply into the tree, child first."""
    new_tree = copy.deepcopy(tree)
    new_children = map(lambda t: deepTransform(function, t), tree.children)
    new_tree.children = new_children
    return function(new_tree)


def makeDeep(f):
    """Turns any function into a "deep" function."""
    def deep_func(tree):
        return deepTransform(f, tree)
    return deep_func


def sift(prop_function, L):
    """Breaks L into two piles: those that the prop_function says are
    true, and those that the prop_function says are false.."""
    return (filter(prop_function, L),
            filter(negate(prop_function), L))


def sublist(l1, l2):
    """Returns 1 if l1 is a sublist of l2, that is, all the elements
    in l1 are in l2.  Otherwise, returns 0.

    Note: this does not take advantage of any ordering of the
    elements."""
    for element in l1:
        if element not in l2: return 0
    return 1
